## Interactive Essential Dynamics -- ver 2.0

## Copyright (C) 2003-2004 John Mongan <jmongan [at] mccammon.ucsd.edu>

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.

## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.

## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import molecule,vmdcallbacks,vmdnumpy,trans,animate,graphics
import Molecule
import Tkinter as Tk
import tkFileDialog,tkMessageBox
from Numeric import *
import MLab
from LinearAlgebra import *
import re,os

class EigenProj:
    
    pattProj = re.compile('^\s*([-0-9.]+)\s+([-0-9.]+)')
    pattSep = re.compile('[,\s]+')

    def __init__(self, root):
        self.veclist = []
        self.sliders = []
        self.sliderCB = []
        self.entry = None
        self.root = root
        self.evec = Molecule.Molecule()
        self.topol = Molecule.Molecule()

    def loadFromFiles(self, strucfile, vecfile, projfile):
        # Load projections (if any)
        self.prj = self.getProj(projfile)
        if self.prj == None: return False
        # Load vectors
        self.evec.rename("Eigenvectors for "+ strucfile)
        if not self.loadEVec(vecfile, strucfile): return False
        self.prj = self.prj[:,:self.evec.numFrames()-2] 
        # Load molecular topology
        self.topol.load(strucfile)
        self.topol.rename(strucfile)
        if self.topol.numAtoms() != self.evec.numAtoms():
            tkMessageBox.showerror("IED", "Topology and eigenvector files must have equal numbers of atoms")
            self.topol.delete()
            self.evec.delete()
            return False
        return True

    def loadFromTraj(self, mol, numVec):
        if molecule.numframes(mol) < 3 * molecule.numatoms(mol):
            tkMessageBox.showerror("IED", "Trajectory must have at least three times as many frames as atoms")
            self.topol.delete()
            self.evec.delete()
            return False
        coords = array([ vmdnumpy.timestep(mol,i) for i in range(molecule.numframes(mol)) ], 'f')
        avgcoords = MLab.mean(coords, axis=0)
        covar = MLab.cov(coords,bias=1)
        eval, evec = eigenvectors(covar)
        coords = coords - avgcoords
        self.prj = matrixmultiply(coords, transpose(evec[:numVec]))/10.0
        
        # Ugly hack because VMD has no way to set num of atoms or copy a molecule
        tfile = os.tempnam()
        molecule.write(mol, 'pdb', tfile, beg = 0, end = 0, waitfor = -1)
        self.evec.load(tfile)
        self.topol.load(tfile)
        os.unlink(tfile)

        self.evec.rename("Eigenvectors for " + molecule.name(mol))
        self.topol.rename("Projection of " + molecule.name(mol))

        self.makeFrames(self.evec, numVec+2)
        put(vmdnumpy.timestep(self.evec, 1), arange(len(avgcoords)),avgcoords)
        for i in range(numVec):
            put(vmdnumpy.timestep(self.evec, i+2), arange(len(evec[i])),evec[i] *10.0)
        return True

                              

    def start(self):
        if (len(self.prj) > 1):
            self.prjLim = [(min(v),max(v)) for v in transpose(self.prj)]
        else:
            self.prjLim = [(-5,5) for v in transpose(self.prj)]
        self.launchGUI()
        self.toplevel.protocol("WM_DELETE_WINDOW", self.destroy)
        self.toplevel.bind("<Destroy>", self.destroy)
        trans.show(self.evec, False)
        animate.activate(self.evec, False)
        self.makeFrames(self.topol,len(self.prj))
        self.topol.setFrame(0)
        self.doProjections()
        trans.resetview(self.topol)

    def destroy(self, event=None):
        self.drawVecCB.deselect()       # To delete associated callback
        vmdcallbacks.del_callback('frame', self.moveSliders)
        self.evec.delete()
        self.topol.delete()
        self.toplevel.unbind("<Destroy>") # Unbind to avoid messages as children are destroyed
                                          # (I know, it's so cruel...)
        if (event == None):             # When closed by window manager, destroy Tk objects
            self.toplevel.destroy()

    def getProj(self,filename):         # Parse projections 
        if filename == '':
            projections = array([ [0 for v in range(30)] ],'f') # Max number of normal modes to visualize
        else:
            projections = []
            file = open(filename, 'r')
            base, ext = os.path.splitext(filename.lower())
            if (ext == '.xvg'):
                vecNum = -1
                timeZero = None
                for line in file:
                    m = EigenProj.pattProj.match(line)
                    if m:
                        time, proj = m.group(1,2)
                        if timeZero == None:
                            timeZero = time
                        if timeZero == time:
                            vecNum = vecNum + 1
                            projections.append( [] )
                        projections[vecNum].append(float(proj))
                projections = transpose(array(projections, 'f'))
            elif (ext == '.ppj'):
                lineNum = 0
                for line in file:
                    if lineNum < 2:
                        lineNum += 1
                        continue
                    else:
                        projs = EigenProj.pattSep.split(line.strip())
                        projections.append( [ float(x) for x in projs[1:] ] )
                projections = array(projections, 'f')        
                
            else:
                print "Can't handle this projection file type"
                return None
        return projections

    def loadEVec(self, vecfile, strucfile):
        base, ext = os.path.splitext(vecfile.lower())
        if ext == '.trr':
            self.evec.load(vecfile,last=len(self.prj[0])+1,waitfor=-1)
            return True
        elif ext == '.pev':
            skipLines = 2
            evecList = []
            evec = []
            file = open(vecfile, 'r')
            for line in file:
                if skipLines > 0:       # Skip headers, record separators
                    skipLines -= 1
                    continue
                if line[1] == '*':
                    skipLines = 1       # Skip evec num and evalue
                    evecList.append(evec)
                    evec = []
                    continue
                elems = [float(x) for x in EigenProj.pattSep.split(line.strip())]
                evec.extend(elems)
            evecList.append(evec)       # Append last eigenvector (no separator at EOF)
            self.evec.load(strucfile)
            self.makeFrames(self.evec, len(evecList)+2)
            for i in range(len(evecList)):
                put(vmdnumpy.timestep(self.evec, i+1), arange(len(evecList[i])), evecList[i])
            return True
        else:
            print "Can't handle this eigenvector file type"
            return False
        
    def makeFrames(self,mol,num):
        for i in range(num-mol.numFrames()):
            mol.dupFrame()

    def doProjections(self):
        activeVecs = [v for v in self.veclist if v.get() >= 0]
        selVecs = transpose(array([(vmdnumpy.timestep(self.evec, v.get()+2))
                         for v in activeVecs]))
        selPrj = take(self.prj, array([v.get() for v in activeVecs]), 1)
        for i in range(len(self.prj)):
            coords = vmdnumpy.timestep(self.topol, i)
            if len(activeVecs) > 0:
                # Add translations along selected eigenvectors
                put(coords, arange(len(coords)),( vmdnumpy.timestep(self.evec,1) 
                                                  + matrixmultiply(selVecs, selPrj[i])
                                                  ))
            else :
                put(coords, arange(len(coords)), vmdnumpy.timestep(self.evec,1))
 
        molecule.set_frame(self.topol,molecule.get_frame(self.topol)) # Force redraw
        
    def launchGUI(self):
        self.toplevel = toplevel = Tk.Toplevel(self.root)
        toplevel.title("IED: "+self.topol.name())
        sliderFrame = ScrolledFrame(toplevel)
        for i in range(len(self.prj[0])):
            fr = Tk.Frame(sliderFrame.inner)
            cbVar = Tk.IntVar()
            cb = Tk.Checkbutton(fr, text = "Eigenvector "+str(i), variable=cbVar,
                                onvalue = i, offvalue=-1)
            cb.pack(side = Tk.LEFT)
            cb.deselect()
            self.veclist.append(cbVar)
            scVar = Tk.DoubleVar()
            scale = Tk.Scale(fr, variable = scVar, from_ = self.prjLim[i][0],
                             to = self.prjLim[i][1], orient=Tk.HORIZONTAL, resolution=0.01
                             )
            scale.pack(side = Tk.RIGHT, expand=True,fill=Tk.X)
            scaleCB = ScaleMove(self,scale,i)
            scale.config(state=Tk.DISABLED, fg='#555555',command=scaleCB)
            self.sliders.append(scale)
            self.sliderCB.append(scaleCB)
            cb.config(command=UpdateScale(cbVar,scale,self))
                      
            fr.pack(expand=True,fill=Tk.BOTH,padx=5)
        sliderFrame.pack(expand=1, fill=Tk.BOTH)
        fr = Tk.Frame(toplevel)
        cbVar = Tk.IntVar()
        self.drawVecCB = Tk.Checkbutton(fr, text = "Draw eigenvector on atom click", variable=cbVar,
                            command=DrawVectors(cbVar,self))
        self.drawVecCB.pack()
        self.entry = Tk.Entry(fr)
        self.entry.config(bg='#FFFFFF')
        self.entry.pack(side=Tk.LEFT)
        pb = Tk.Button(fr, text="Set Projections", command=self.setProjections)
        pb.pack(side=Tk.RIGHT)
        fr.pack(padx=5,pady=10)
        vmdcallbacks.add_callback('frame', self.moveSliders)


        
    def moveSliders(self, molid, frame):
        for i in range(len(self.prj[0])):
            if self.veclist[i].get() != -1:
                self.sliderCB[i].ignore = True
                self.sliders[i].set(self.prj[frame][i])

    def setProjections(self):
        projs = EigenProj.pattSep.split(self.entry.get())
        self.entry.delete(0,len(self.entry.get()))
        for i in range(min(len(self.sliders),len(projs))):
            self.sliders[i].set(float(projs[i]))
        
class DrawVectors:                      # Respond to "Draw eigenvector" CB
    def __init__(self, cbVar, eigenProj):
        self.eigenProj = eigenProj
        self.cbVar = cbVar
    def __call__(self):
        if self.cbVar.get():
            vecs = [v for v in self.eigenProj.veclist if v.get() >= 0]
            if len(vecs) == 1:          # Check that only one vector is selected
                self.vecNum = vecs[0].get()
                vmdcallbacks.add_callback('pick_atom', self.placeVector)
                graphics.color(self.eigenProj.topol, 'silver')
            else:
                tkMessageBox.showerror("IED", "You must select exactly one eigenvector before enabling "+
                                       "vector drawing.")
                self.eigenProj.drawVecCB.deselect()
        else:
            vmdcallbacks.del_callback('pick_atom', self.placeVector)

    def placeVector(self,molid, atomid, key_shift_state):
        avgLoc = vmdnumpy.timestep(self.eigenProj.evec, 1)[atomid*3:atomid*3 +3]
        vec = vmdnumpy.timestep(self.eigenProj.evec,self.vecNum+2)[atomid*3:atomid*3 +3]
        headLoc = avgLoc + vec*self.eigenProj.prjLim[self.vecNum][1]
        tailLoc = avgLoc + vec*self.eigenProj.prjLim[self.vecNum][0]
        drawArrow(self.eigenProj.topol, tailLoc, headLoc)
        
def drawArrow(mol, start, end):
    headStart = (end - start) * 0.66 + start
    graphics.cylinder(mol, tuple(start), tuple(headStart), radius = 0.1, filled=1)
    graphics.cone(mol, tuple(headStart), tuple(end), radius = 0.2)
   
class ScaleMove:
    def __init__(self, eigenProj, scale, vecNum):
        self.eigenProj = eigenProj
        self.scale = scale
        self.vecNum = vecNum
        self.ignore = False             # When true, next event caused by frame change
    def __call__(self, dummyArg):
        if self.ignore:                 # Program moved slider b/c of frame change
            self.ignore = False
        else :                          # User moved slider -> update coordinates
            selVecs = transpose(array([(vmdnumpy.timestep(self.eigenProj.evec, v.get()+2))
                                       for v in self.eigenProj.veclist if v.get() >= 0]))
            prj = [self.eigenProj.sliders[i].get() for i in range(len(self.eigenProj.sliders)) if self.eigenProj.veclist[i].get()>=0]
            frame = self.eigenProj.topol.curFrame()
            coords = vmdnumpy.timestep(self.eigenProj.topol, frame)
            put(coords, arange(len(coords)),( vmdnumpy.timestep(self.eigenProj.evec,1) 
                                              + matrixmultiply(selVecs, prj)
                                              ))
            self.eigenProj.topol.setFrame(self.eigenProj.topol.curFrame())

class UpdateScale:
    def __init__(self, cbVar, scale, ep):
        self.cbVar = cbVar
        self.scale = scale
        self.ep = ep

    def __call__(self):
        graphics.delete(self.ep.topol, 'all')
        self.ep.drawVecCB.deselect()
        if self.cbVar.get() == -1 :
            self.scale.set(0)
            self.scale.config(state=Tk.DISABLED, fg='#555555')
        else :
            self.scale.config(state=Tk.NORMAL, fg='#000000')
        self.ep.moveSliders(self.ep.topol, self.ep.topol.curFrame())
        self.ep.doProjections()


class EPLoader:
    def __init__(self):
        self.root = None
        self.launchGUI()
        
    def launchGUI(self):
        self.root = Tk.Tk()
        self.root.title("Interactive Essential Dynamics")
        self.root.protocol("WM_DELETE_WINDOW", self.root.destroy)
        pb = Tk.Button(self.root, text = "Load ED data from files", command = self.doEPFileLoader)
        pb.pack(padx=15,pady=15)
        pb = Tk.Button(self.root, text = "Calculate ED from trajectory", command = self.doEPTrajLoader)
        pb.pack(padx=15,pady=15)

    def doEPFileLoader(self):
        EPFileLoader(self.root)

    def doEPTrajLoader(self):
        EPTrajLoader(self.root)

class EPTrajLoader:
    def __init__(self, root):
        self.root = root
        self.launchGUI()

    def launchGUI(self):
        self.win = Tk.Toplevel(self.root)
        self.win.title("IED: Calculate ED")
        label = Tk.Label(self.win, text="Trajectory ID")
        label.pack()
        self.tidEntry = Tk.Entry(self.win)
        self.tidEntry.pack()
        self.tidEntry.insert(Tk.INSERT,str(molecule.get_top()))
        label = Tk.Label(self.win, text = "Number of Eigenvectors to display")
        label.pack()
        self.evnEntry = Tk.Entry(self.win)
        self.evnEntry.pack()
        self.evnEntry.insert(Tk.INSERT,"10")
        
        pb = Tk.Button(self.win, text="Calculate", command=self.loadTraj) 
        pb.pack(side=Tk.TOP,pady=15)

    def validateEntries(self):
        tidstr = self.tidEntry.get()
        evnstr = self.evnEntry.get()
        try:
            tid = int(tidstr)
            evn = int(evnstr)
        except ValueError:
            tkMessageBox.showerror("Entry error", "Trajectory ID and Eigenvector number must be numeric values")
            return False
        if (not tid in molecule.listall()):
            tkMessageBox.showerror("Entry error", str(tid) + " is not a valid trajectory ID")
            return False
        maxVec = 3 * molecule.numatoms(tid)
        if (evn < 0 or evn > maxVec):
            tkMessageBox.showerror("Entry error", "Number of Eigenvectors must be between 1 and "+str(maxVec))
            return False
        return True
        

    def loadTraj(self):
        if (self.validateEntries() and tkMessageBox.askyesno("ED Calculation", "Warning:\n" +
                                      "You must \"fit\" (remove rotations and translations) your trajectory before " +
                                      "proceeding.\n" +                       
                                      "Depending on the number of atoms in your molecule, the ED calculation may take a long " +
                                      "time (minutes to hours).\nThe VMD display will freeze (appear to have crashed) "+
                                      "until the calculation is complete. Continue?")):
            ep = EigenProj(self.root)
            if (ep.loadFromTraj(int(self.tidEntry.get()),int(self.evnEntry.get()))):
                trans.show(int(self.tidEntry.get()), False)
                self.win.destroy()
                ep.start()


class EPFileLoader:
    def __init__(self, root):
        self.label = ("Topology file", "Eigenvector file", "Projections file")
        self.ext = [ [("PDB file","*.pdb"),("All files", "*")],
                     [("GROMACS eigenvectors","*.trr"),("Ptraj eigenvectors", "*.pev")],
                     [("GROMACS projections","*.xvg"),("Ptraj projections","*.ppj")]
                   ]
        self.entry = []
        self.default = ('stopol.pdb','eigenvec.trr','proj.xvg')
        self.root = root
        self.launchGUI()

    def launchGUI(self):
        self.win = Tk.Toplevel(self.root)
        self.win.title("IED: Load ED Data")
        for i in range(len(self.label)):
            fr = Tk.Frame(self.win)
            label = Tk.Label(fr, text = self.label[i])
            label.grid(row=0,column=0,sticky=Tk.W)
            self.entry.append(Tk.Entry(fr))
            self.entry[i].config(bg='#FFFFFF')
            self.entry[i].insert(0,self.default[i])
            self.entry[i].grid(row=1,column=0)
            pb = Tk.Button(fr, text="Browse...", command=BrowseCB(self.entry[i],self.label[i],
                                                                  self.ext[i],self.win))
            pb.grid(row=1,column=1)
            fr.pack(side=Tk.TOP,padx=15,pady=5)
        pb = Tk.Button(self.win, text="Load", command=self.loadFiles)
        pb.pack(side=Tk.TOP,pady=15)
                           
    def loadFiles(self):
        for i in range(len(self.label)):
            if (not os.access(self.entry[i].get(),os.R_OK) and not(i==2 and self.entry[i].get()=='')):
                tkMessageBox.showerror("File error",
                                       "Can't read file " + self.entry[i].get())
                return
        ep = EigenProj(self.root)
        if (ep.loadFromFiles(self.entry[0].get(),self.entry[1].get(), self.entry[2].get())):
            self.win.destroy()
            ep.start()

class BrowseCB:
    def __init__(self, entry, label, ext, parent):
        self.entry = entry
        self.label = label
        self.ext = ext
        self.parent = parent

    def __call__(self):
        filename = tkFileDialog.askopenfilename(parent=self.parent,
                                                title = self.label,
                                                filetypes = self.ext)
        if (filename != ''):
            self.entry.delete(0,len(self.entry.get()))
            self.entry.insert(0,filename)

# scrolled frame
#
# Adapted and expanded from work posted by
#
# fredrik lundh, april 1998
#
# fredrik@pythonware.com
# http://www.pythonware.com
#

class ScrolledFrame(Tk.Frame):

    def __init__(self, master, **kw):
        apply(Tk.Frame.__init__, (self, master), kw)

        # create the widgets
        scrollbar = Tk.Scrollbar(self, orient=Tk.VERTICAL)
        self.scrollbar = scrollbar
        self.scrollbarVis = True

        self.canvas = Tk.Canvas(
            self,
            bd=0, highlightthickness=0,
            yscrollcommand=scrollbar.set
            )

        scrollbar.config(command=self.canvas.yview)

        scrollbar.pack(fill=Tk.Y, side=Tk.RIGHT)
        self.canvas.pack(expand=1, fill=Tk.BOTH, side=Tk.LEFT, )

        # reset the view
        # (always do this if you don't use scrollbars)
        self.canvas.xview("moveto", 0)
        self.canvas.yview("moveto", 0)

        # create the inner frame
        self.inner = Tk.Frame(self.canvas)

        # track changes to its size
        self.inner.bind('<Configure>', self.__configure)
        self.canvas.bind('<Configure>', self.__canvasResize)
 
        # place the frame inside the canvas (this also
        # runs the __configure method)
        self.frameWin = self.canvas.create_window(0, 0, window=self.inner, anchor=Tk.NW)
    def __canvasResize(self, event):
        if self.canvas.winfo_height() > self.inner.winfo_reqheight():
            self.canvas.itemconfigure(self.frameWin,height=self.canvas.winfo_height())
            if self.scrollbarVis == True :
                self.scrollbar.pack_forget()
                self.scrollbarVis = False
        else:
            self.canvas.itemconfigure(self.frameWin,height=self.inner.winfo_reqheight())
            if self.scrollbarVis == False :
                self.canvas.pack_forget() # Repack so scrollbar has priority for horizonal space
                self.scrollbar.pack(fill=Tk.Y, side=Tk.RIGHT)
                self.canvas.pack(expand=1, fill=Tk.BOTH, side=Tk.LEFT, )
                self.scrollbarVis = True
        if self.canvas.winfo_width() > self.inner.winfo_reqwidth():
            self.canvas.itemconfigure(self.frameWin,width=self.canvas.winfo_width())
        else:
            self.canvas.itemconfigure(self.frameWin,width=self.inner.winfo_reqwidth())
       

    def __configure(self, event):

        # update the scrollbars to match the size of the inner frame
        size = self.inner.winfo_reqwidth(), self.inner.winfo_reqheight()
        self.canvas.config(scrollregion="0 0 %s %s" % size)


def tkmenu():
    print ""
    print "Interactive Essential Dynamics  v 2.0 (http://mccammon.ucsd.edu/ied)"
    print "Copyright John Mongan <jmongan@mccamon.ucsd.edu>, 2003, 2004"
    print ""
    print "Citation: Interactive Essential Dynamics. John Mongan. Journal of"
    print "          Computer-Aided Molecular Design. In press."
    print ""
    print "This library is provided under the LGPL with ABSOLUTELY NO WARRANTY."
    print ""
    return EPLoader().root

if __name__=="__main__":
    tkmenu()

